/*

    File: Main.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>		//this will go after i find a better place to init opengl vars
#include <math.h>

#include "../m_system3d/main.h"
#include "../m_foundation/main.h"
#include "../m_foundation/random.h"
#include "../m_foundation/const.h"

/*

  realtime cloth:
gridLevel = 16
dt = 0.5
intType = 1			//RK4
ks = 1
kd = 0.1
env = 0.04

  high-quality realistic slow cloth
gridLevel = 64
dt = 0.1			//or -1
intType = 0
ks = 1
kd = 0.1
env = 0.1
*/


Texture worldTex;

typedef struct vertex_s {
	//system-based
	bool		active;
	scalar_t	mass_inverse;
	scalar_t	uv[2];
	vector_t	normal;
	//physics
	vector_t	position;
	vector_t	velocity;
	vector_t	force;
	//RK4 dummy vars - only 120 extra bytes per vertex. EGADS!
	vector_t	source_pos, source_vel;
	vector_t	tempvel[4], tempforce[4];
} vertex_t;

typedef struct spring_s {
	bool		active;
	vertex_t	*v[2];		//vertex children
	scalar_t	restLen;	//rest length
} spring_t;

#define GRID_MAX	64

//#define SPRING_MAX		(6*GRID*GRID - 10*GRID + 2)
//#define SPRING_MAX	(2*GRID*GRID - 2*GRID + 2*GRID*GRID - 4*GRID + 2 + 2*GRID*GRID - 4*GRID)
//#define SPRING_MAX	(2*GRID*GRID-2*GRID+2*GRID*GRID-4*GRID+2+2*GRID*GRID-4*GRID)
//#define SPRING_MAX	(GRID*(GRID-1)+(GRID-1)*GRID+2*(GRID-1)*(GRID-1)+2*(GRID-2)*GRID)
//#define VERTEX_MAX		(GRID*GRID)

const int springMax = 6 * GRID_MAX * GRID_MAX - 10 * GRID_MAX + 2;
const int vertexMax = GRID_MAX * GRID_MAX;

int gridLevel = GRID_MAX;
int springCount = springMax;
int vertexCount = vertexMax;

spring_t spring[springMax];
vertex_t vertex[vertexMax];

spring_t *CreateSpring(vertex_t *v0, vertex_t *v1)
{
	int i;

	spring_t *sp = spring;

	for (i = 0; i < springCount; i++)
	{
		sp = &spring[i];
		if (!sp->active) break;
	}

	sp->active = 1;
	sp->v[0] = v0;
	sp->v[1] = v1;
	sp->restLen = VectorDist(&v0->position, &v1->position);

	return sp;
}

void CFunc_Reset(int argc,char **argv)
{
	int i,j;

	if (argc > 1)
	{
		int newGrid = atoi(argv[1]);
		if (newGrid > 0 && newGrid <= GRID_MAX)
		{
			gridLevel = newGrid;
			springCount = 6 * gridLevel * gridLevel - 10 * gridLevel + 2;
			vertexCount = gridLevel * gridLevel;
		}
	}

	for (i = 0; i < springCount; i++)
		memset(&spring[i], 0, sizeof(spring_t));
	for (i = 0; i < vertexCount; i++)
		memset(&vertex[i], 0, sizeof(vertex_t));

	for (i = 0; i < gridLevel; i++)
	{
		for (j = 0; j < gridLevel; j++)
		{
			vertex[i+j*gridLevel].active = 1;
			vertex[i+j*gridLevel].mass_inverse = 1;
			vertex[i+j*gridLevel].uv[0] = (scalar_t)i/(gridLevel-1);
			vertex[i+j*gridLevel].uv[1] = (scalar_t)j/(gridLevel-1);
#if 1	//linear
			vertex[i+j*gridLevel].position.x = 0;//-((vertex[i+j*gridLevel].uv[0]-0.5f)*(vertex[i+j*gridLevel].uv[0]-0.5f)+(vertex[i+j*gridLevel].uv[1]-0.5f)*(vertex[i+j*gridLevel].uv[1]-0.5f))*200.f;
			vertex[i+j*gridLevel].position.y = -vertex[i+j*gridLevel].uv[0]*200.f;
			vertex[i+j*gridLevel].position.z = -vertex[i+j*gridLevel].uv[1]*200.f;
		}
	}

	for (i = 0; i < gridLevel; i++)
	{
		for (j = 0; j < gridLevel; j++)
		{
			if (i > 0) CreateSpring(&vertex[i+j*gridLevel], &vertex[i-1+j*gridLevel]);
			if (j > 0) CreateSpring(&vertex[i+j*gridLevel], &vertex[i+(j-1)*gridLevel]);
			if (i > 0 && j > 0)
				CreateSpring(&vertex[i+j*gridLevel], &vertex[i-1+(j-1)*gridLevel]);
			if (i > 0 && j < gridLevel-1)
				CreateSpring(&vertex[i+j*gridLevel], &vertex[i-1+(j+1)*gridLevel]);

			if (i > 1) CreateSpring(&vertex[i+j*gridLevel], &vertex[i-2+j*gridLevel]);
			if (j > 1) CreateSpring(&vertex[i+j*gridLevel], &vertex[i+(j-2)*gridLevel]);
#endif
#if 0		//sphere
			VectorSet(&vertex[i+j*gridLevel].position, 0, 0, 200);
			quat_t blah;
			AngleAxisToQuat(0,1,0,(scalar_t)i/(scalar_t)(gridLevel-1)*160.f + 10.f, &blah);
			QuatRotateVector(&vertex[i+j*gridLevel].position, &blah, &vertex[i+j*gridLevel].position);
			AngleAxisToQuat(0,0,1,(scalar_t)j/(scalar_t)(gridLevel-1)*340.f + 10.f, &blah);
			QuatRotateVector(&vertex[i+j*gridLevel].position, &blah, &vertex[i+j*gridLevel].position);
		}
	}

	for (i = 0; i < gridLevel; i++)
	{
		for (j = 0; j < gridLevel; j++)
		{
			int di = (i + gridLevel - 1) % gridLevel;
			int ddi = (di + gridLevel - 1) % gridLevel;
			int nj = (j + 1) % gridLevel;
			int dj = (j + gridLevel - 1) % gridLevel;
			int ddj = (dj + gridLevel - 1) % gridLevel;
			CreateSpring(&vertex[i+j*gridLevel], &vertex[di+j*gridLevel]);
			CreateSpring(&vertex[i+j*gridLevel], &vertex[i+dj*gridLevel]);
			CreateSpring(&vertex[i+j*gridLevel], &vertex[di+dj*gridLevel]);
			CreateSpring(&vertex[i+j*gridLevel], &vertex[di+nj*gridLevel]);
			CreateSpring(&vertex[i+j*gridLevel], &vertex[ddi+j*gridLevel]);
			CreateSpring(&vertex[i+j*gridLevel], &vertex[i+ddj*gridLevel]);
#endif
		}
	}
}

// ***************************** DIRECT HACK COPY FROM WIREFRAME VIS TEST

//lighting vars
static int useLighting = 1;
static quat_t LightPos = {1,1,0,0};
static int useSeparateSpec = 1;
static float materialShininess = 16.f;
static quat_t materialAmbient = {0.25f,0.25f,0.25f,1};
static quat_t materialDiffuse = {1,1,1,1};
static quat_t materialSpecular = {0.5f,0.5f,0.5f,1};

void CFunc_updateLighting(int argc, char **argv)
{
	static quat_t lq0001 = {0,0,0,1};
	static quat_t lq1111 = {1,1,1,1};
	
	glLightfv(GL_LIGHT0, GL_AMBIENT, lq1111.v);				//keep this at full - edit ambient light color through material ambient
	glEnable(GL_LIGHT0);
	
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER , GL_FALSE);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lq0001.v);

	if (found_GL_EXT_separate_specular_color)
	{
		if (useSeparateSpec)
			glLightModelf(GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SEPARATE_SPECULAR_COLOR_EXT);
		else
			glLightModelf(GL_LIGHT_MODEL_COLOR_CONTROL_EXT, GL_SINGLE_COLOR_EXT);
	}

	float max_shininess = 128;
	if (found_GL_NV_light_max_exponent)
		glGetFloatv(GL_MAX_SHININESS_NV, &max_shininess);


	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, materialAmbient.v);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, materialDiffuse.v);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, materialSpecular.v);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, materialShininess);

	glEnable(GL_NORMALIZE);						//because its hardware accelerated!
}

// ***************************** END DIRECT HACK COPY 

static scalar_t timestep = 0.5f;//-1;//0.1f;		//negative implies "use frame timestep"
static scalar_t env_damping = 0.04f;
static scalar_t world_ks = 1.f;
static scalar_t world_kd = 0.1f;
static vector_t gravity = {0,0,0};
static scalar_t tex_scalar = 1.f;

bool Init_Application3D(void)
{
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	CVar_Exec("ConsoleFont font.bmp");
	CVar_Exec("Console idlevis 0");
	CVar_Exec("cameraKeys 1");
	CVar_Exec("camSpeed 3");
	CVar_Exec("camRot 3");
	CVar_Exec("CamAngle 0 0 1 0");
	CVar_Exec("CamPos 150 -100 -100");


	worldTex.Create("texture.bmp");

	new CVarw(&timestep, "dt");
	new CVarw(&env_damping, "env");
	new CVarw(&world_ks, "ks");
	new CVarw(&world_kd, "kd");
	new CVarw(&gravity, "grav");
	new CVarw(&tex_scalar, "tex_scalar");
	new CVarw(0,	"reset",	CFunc_Reset);

	CVar_Exec("reset 16");

	glDisable(GL_CULL_FACE);

// ***************************** DIRECT COPY FROM WIREFRAME VIS TEST

	new CVarw(&useLighting,		"useLighting");
		//lighting based
	new CVarw(&LightPos,			"LightPos");
	new CVarw(&useSeparateSpec,	"useSeparateSpec",	CFunc_updateLighting);
	new CVarw(&materialShininess,	"materialShininess",CFunc_updateLighting);
	new CVarw(&materialAmbient,	"materialAmbient",	CFunc_updateLighting);
	new CVarw(&materialDiffuse,	"materialDiffuse",	CFunc_updateLighting);
	new CVarw(&materialSpecular,	"materialSpecular",	CFunc_updateLighting);
		//UPDATE IT ALL FOR THE FIRST ROUND
	CFunc_updateLighting(0,0);

// ***************************** END DIRECT COPY 

	return true;
}

vertex_t *mousevtx = NULL;

void ComputeForces(void)
{
	spring_t *sp;
	vertex_t *vtx;

	vector_t deltaP, deltaV, springForce;
	scalar_t dist, H, D, sum;
	int i;

//clear all forces
	for (i = 0; i < vertexCount; i++)
	{
		vtx = &vertex[i];
		if (!vtx->active) continue;

		//initialize it to the environment force
		VectorCopy(&gravity, &vtx->force);

		//apply initial damping
		VectorMultAdd(&vtx->force, &vtx->velocity, -env_damping, &vtx->force);
	}
//update mouse force - simulate spring
	if (main_mouse->buttonState[0] == MH_DOWN)
	{
		vector_t mousePos;
		scalar_t influence;

		static scalar_t mouse_force = 10.f;
		static scalar_t influence_alpha_sq = 100.f;

		static CVarw *fsdj[] = {
			new CVarw(&influence_alpha_sq, "alpha_sq"),
			new CVarw(&mouse_force, "mf"),
			NULL};

		VectorSub(&vtx->position, &main_camera->position, &mousePos);
		VectorMultAdd(
			&main_camera->mouseLine.pos,
			&main_camera->mouseLine.dir,
			VectorDot(&main_camera->matrix.vec[0], &mousePos) / main_camera->znear,
			&mousePos);

		//always calc force via mousevtx - therefore all other objs in influence will be pulled in same force in same dir
		VectorSub(&mousevtx->position, &mousePos, &deltaP);
		dist = VectorLength(&deltaP);
		H = dist * world_ks;

		for (i = 0; i < vertexCount; i++)
		{
			vtx = &vertex[i];
			if (!vtx->active) continue;

			influence = (scalar_t)exp(-VectorDistSq(&vtx->position, &mousevtx->position) / influence_alpha_sq);

			D = (VectorDot(&vtx->velocity, &deltaP) * world_kd) / dist;

			VectorScale(&deltaP, -(H + D) * influence / dist, &springForce);

			VectorAdd(&vtx->force, &springForce, &vtx->force);
		}
	}
//update spring forces
	for (i = 0; i < springCount; i++)
	{
		sp = &spring[i];
		if (!sp->active) continue;

		VectorSub(&sp->v[0]->position, &sp->v[1]->position, &deltaP);
		dist = VectorLength(&deltaP);
		H = (dist - sp->restLen) * world_ks;


		VectorSub(&sp->v[0]->velocity, &sp->v[1]->velocity, &deltaV);
		D = (VectorDot(&deltaV, &deltaP) * world_kd) / dist;

		sum = H + D;

		VectorScale(&deltaP, -sum / dist, &springForce);

		VectorAdd(&sp->v[0]->force, &springForce, &sp->v[0]->force);
		VectorSub(&sp->v[1]->force, &springForce, &sp->v[1]->force);
	}
}

void DisplayObject(void)
{
	static int drawSprings = 0;
	static int drawVertexes = 0;
	static int drawTex = 1;
	static CVarw *fdjksfd[] = {
		new CVarw(&drawSprings, "drawSprings"),
		new CVarw(&drawVertexes, "drawVertexes"),
		new CVarw(&drawTex, "drawTex"),
		NULL};

	spring_t *sp;
	vertex_t *vtx;

	vector_t deltaP, deltaV;
	scalar_t dist, H, D, sum;
	vector_t color;
	int i;

//display spring system
	if (drawSprings)
	{
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_LINES);
		glColor3f(1,0,0);
		for (i = 0; i < springCount; i++)
		{
			sp = &spring[i];
			if (!sp->active) continue;

			VectorSub(&sp->v[0]->position, &sp->v[1]->position, &deltaP);
			dist = VectorLength(&deltaP);
			H = (dist - sp->restLen) * world_ks;

			VectorSub(&sp->v[0]->velocity, &sp->v[1]->velocity, &deltaV);
			D = (VectorDot(&deltaV, &deltaP) * world_kd) / dist;

			sum = H + D;

			color.x = sum < 0 ? 1.f + sum / 5.f : 1.f;
			color.y = sum > 0 ? 1.f - sum / 5.f : 1.f;
			color.z = 0.f;
			
			glColor3fv(color.v);
			glVertex3fv(sp->v[0]->position.v);
			glVertex3fv(sp->v[1]->position.v);
		}
		glEnd();
		glEnable(GL_TEXTURE_2D);
	}

//display verticies
	if (drawVertexes)
	{
		glDisable(GL_TEXTURE_2D);
		glPointSize(3);
		glBegin(GL_POINTS);
		for (i = 0; i < vertexCount; i++)
		{
			vtx = &vertex[i];
			if (!vtx->active) continue;

			color.x = (vtx != mousevtx);
			color.y = vtx->mass_inverse;
			color.z = 1;

			glColor3fv(color.v);
			glVertex3fv(vtx->position.v);
		}
		glEnd();
		glEnable(GL_TEXTURE_2D);
	}

	if (drawTex)
	{
		int x,y;
		
		worldTex.Set();
		glColor3f(1,1,1);

		if (useLighting)	//update this every time the modelview matrix changes
		{
			glLightfv(GL_LIGHT0, GL_POSITION, LightPos.v);	//reset position in modelview matrix
			glEnable(GL_LIGHTING);
		}

		i = 0;
		for (y = 0; y < gridLevel-1; y++)
		{
			glBegin(GL_TRIANGLE_STRIP);
			for (x = 0; x < gridLevel; x++)
			{
				i += gridLevel;

				if (useLighting && x < gridLevel-1 && y < gridLevel-2)
				{
					VectorPlaneNormal(
						&vertex[i+1].position,
						&vertex[i].position,
						&vertex[i+gridLevel].position,
						&vertex[i].normal);
					glNormal3fv(vertex[i].normal.v);
				}
				
				glTexCoord2f(vertex[i].uv[0] * tex_scalar, vertex[i].uv[1] * tex_scalar);
				glVertex3fv(vertex[i].position.v);

				i -= gridLevel;

				if (useLighting && x < gridLevel - 1)	//only render the first row this time around
				{
					if (y == 0)
					{
						VectorPlaneNormal(
							&vertex[i+1].position,
							&vertex[i].position,
							&vertex[i+gridLevel].position,
							&vertex[i].normal);
					}

					glNormal3fv(vertex[i].normal.v);
				}
				
				glTexCoord2f(vertex[i].uv[0] * tex_scalar, vertex[i].uv[1] * tex_scalar);
				glVertex3fv(vertex[i].position.v);

				i++;
			}
			glEnd();
		}

		if (useLighting)
		{
			glDisable(GL_LIGHTING);
		}

	}
}

void Integrate_Euler(scalar_t dt)
{
	static int useAxis[3] = {1,1,1};
	static scalar_t max_spring_force = 100.f;
	static CVarw *fdj[] = {
		new CVarw(&useAxis[0], "useX"),
		new CVarw(&useAxis[1], "useY"),
		new CVarw(&useAxis[2], "useZ"),
		new CVarw(&max_spring_force, "max_spring_force"),
		NULL};

	int i,j;
	scalar_t force;
	vertex_t *vtx;

//compute forces
	ComputeForces();

//update verticies
	for (i = 0; i < vertexCount; i++)
	{
		vtx = &vertex[i];
		if (!vtx->active) continue;

		force = VectorLengthSq(&vtx->force);
		if (force > max_spring_force*max_spring_force)
		{
			force = (scalar_t)sqrt(force);
			main_console->Print("found particle with force %f\n", force);
			force = max_spring_force / force;

			vtx->force.x *= force;
			vtx->force.y *= force;
			vtx->force.z *= force;
		}

		for (j = 0; j < 3; j++)
		{
			if (!useAxis[j]) continue;
			vtx->position.v[j] += vtx->velocity.v[j] * dt;
			vtx->velocity.v[j] += vtx->force.v[j] * dt * vtx->mass_inverse;
		}
	}

	//upon movement, test for collision...somewhere in here

}

/*	Runge Kutta 4-step

//begin physics frame

compute_forces(current)

forall(v)
	source_pos	= current.pos
	source_vel	= current.vel

forall(v)
	temp1.force	= dt/2 * current.force * current.massinv
	temp1.vel	= dt/2 * source_vel
	current.vel	= source_vel + temp1.force
	current.massinv	= current.massinv
	current.pos	= source_pos + temp1.vel

compute_forces(current)

forall(v)
	temp2.force	= dt/2 * current.force * current.massinv
	temp2.vel	= dt/2 * current.vel
	current.vel	= source_vel + temp2.force
	current.massinv	= current.massinv
	current.pos	= source_pos + temp2.vel

compute_forces(current)

forall(v)
	temp3.force	= dt * current.force * current.massinv
	temp3.vel	= dt * current.vel
	current.vel	= source_vel + temp3.force
	current.massinv	= current.massinv
	current.pos	= source_pos + temp3.vel

compute_forces(current)

forall(v)
	temp4.force	= dt * current.force * current.massinv
	temp4.vel	= dt * current.vel

forall(v)
	current.pos	= source_pos + dt/6 * (temp1.vel + 2 * (temp2.vel + temp3.vel) + temp4.vel)
	current.vel	= source_vel + dt/6 * (temp1.force + 2 * (temp2.force + temp3.force) + temp4.force)

//end physics frame

*/
void Integrate_RK4(scalar_t dt)
{
	int i;
	vertex_t *vtx;
	scalar_t half_dt = dt / 2.f;
	scalar_t sixth_dt = 1.f / 6.f;
//begin physics frame

	ComputeForces();

	for (i = 0; i < vertexCount; i++)
	{
		vtx = &vertex[i];
		if (!vtx->active) continue;

		//backup frame source states
		VectorCopy(&vtx->position, &vtx->source_pos);
		VectorCopy(&vtx->velocity, &vtx->source_vel);

		//compute first temp system
		VectorScale(&vtx->force, vtx->mass_inverse * half_dt, &vtx->tempforce[0]);
		VectorScale(&vtx->source_vel, half_dt, &vtx->tempvel[0]);

		//integrate system
		VectorAdd(&vtx->source_vel, &vtx->tempforce[0], &vtx->velocity);
		VectorAdd(&vtx->source_pos, &vtx->tempvel[0], &vtx->position);

	}

	ComputeForces();

	for (i = 0; i < vertexCount; i++)
	{
		vtx = &vertex[i];
		if (!vtx->active) continue;

		//compute second temp system
		VectorScale(&vtx->force, vtx->mass_inverse * half_dt, &vtx->tempforce[1]);
		VectorScale(&vtx->velocity, half_dt, &vtx->tempvel[1]);

		//integrate system
		VectorAdd(&vtx->source_vel, &vtx->tempforce[1], &vtx->velocity);
		VectorAdd(&vtx->source_pos, &vtx->tempvel[1], &vtx->position);
	}

	ComputeForces();

	for (i = 0; i < vertexCount; i++)
	{
		vtx = &vertex[i];
		if (!vtx->active) continue;

		//compute third temp system
		VectorScale(&vtx->force, vtx->mass_inverse * dt, &vtx->tempforce[2]);
		VectorScale(&vtx->velocity, dt, &vtx->tempvel[2]);

		//integrate system
		VectorAdd(&vtx->source_vel, &vtx->tempforce[2], &vtx->velocity);
		VectorAdd(&vtx->source_pos, &vtx->tempvel[2], &vtx->position);
	}

	ComputeForces();

	for (i = 0; i < vertexCount; i++)
	{
		vtx = &vertex[i];
		if (!vtx->active) continue;

		//compute forth temp system
		VectorScale(&vtx->force, vtx->mass_inverse * dt, &vtx->tempforce[3]);
		VectorScale(&vtx->velocity, dt, &vtx->tempvel[3]);

		//and do that final integration like it aint no thang!
		vtx->position.x = vtx->source_pos.x + sixth_dt * (vtx->tempvel[0].x + 2.f * (vtx->tempvel[1].x + vtx->tempvel[2].x) + vtx->tempvel[3].x);
		vtx->position.y = vtx->source_pos.y + sixth_dt * (vtx->tempvel[0].y + 2.f * (vtx->tempvel[1].y + vtx->tempvel[2].y) + vtx->tempvel[3].y);
		vtx->position.z = vtx->source_pos.z + sixth_dt * (vtx->tempvel[0].z + 2.f * (vtx->tempvel[1].z + vtx->tempvel[2].z) + vtx->tempvel[3].z);
		
		vtx->velocity.x = vtx->source_vel.x + sixth_dt * (vtx->tempforce[0].x + 2.f * (vtx->tempforce[1].x + vtx->tempforce[2].x) + vtx->tempforce[3].x);
		vtx->velocity.y = vtx->source_vel.y + sixth_dt * (vtx->tempforce[0].y + 2.f * (vtx->tempforce[1].y + vtx->tempforce[2].y) + vtx->tempforce[3].y);
		vtx->velocity.z = vtx->source_vel.z + sixth_dt * (vtx->tempforce[0].z + 2.f * (vtx->tempforce[1].z + vtx->tempforce[2].z) + vtx->tempforce[3].z);
	}

//end physics frame
}

void Update_Application3D(void)
{
	int i;
	spring_t *sp = NULL;
	vertex_t *vtx = NULL;
	scalar_t dt, dist;

	static double lasttime = 0;
	double thistime = main_timer->GetTime();
	double deltatime = thistime - lasttime;
	lasttime = thistime;

	static int integrationType = 1;
	static CVarw *fdj[] = {
		new CVarw(&integrationType, "intType"),
		NULL};

	dt = timestep;
	if (dt < 0) dt = (scalar_t)deltatime;

	switch (integrationType)
	{
	case 0:
		Integrate_Euler(dt);
		break;
	case 1:
		Integrate_RK4(dt);
		break;
	}

//choose mouse vertex
	if (main_mouse->buttonState[0] == MH_UP)
	{
		scalar_t bestdist = fInf;
		for (i = 0; i < vertexCount; i++)
		{
			vtx = &vertex[i];
			if (!vtx->active) continue;

			//if the mouse has been released - we are valid to check for an input vertex
			dist = LineVectorDistance(&main_camera->mouseLine, &vtx->position);
			if (dist < bestdist)
			{
				bestdist = dist;
				mousevtx = vtx;
			}
		}
	}
	if (mousevtx)
	{
		static scalar_t select = 1;
		if (main_mouse->buttonState[1] == MH_PRESS)
		{
			select = (scalar_t)!(int)mousevtx->mass_inverse;
		}
		if (main_mouse->buttonState[1] == MH_DOWN)
		{
			mousevtx->mass_inverse = select;
			VectorSet(&mousevtx->velocity, 0, 0, 0);
			VectorSet(&mousevtx->force, 0, 0, 0);
		}
	}

	DisplayObject();
}


void Shutdown_Application3D(void)
{
	worldTex.Destroy();
}
